@@ -30,10 +30,10 @@ class DoublyRobustEstimator:
     were never taken by the exploration policy, as it cannot construct a model for them.
     
     The expected rewards are estimated with the imputer algorithm passed here, which should
-    output a number in the range [0,1].
+    output a number in the range :math:`[0,1]`.
     
-    This technique is meant for the case of contiunous rewards in the [0,1] interval,
-    but here it is used for the case of discrete rewards {0,1}, under which it performs
+    This technique is meant for the case of contiunous rewards in the :math:`[0,1]` interval,
+    but here it is used for the case of discrete rewards :math:`\{0,1\}`, under which it performs
     poorly. It is not recommended to use, but provided for comparison purposes.
     
     
@@ -50,7 +50,8 @@ class DoublyRobustEstimator:
     they were chosen they resulted in a reward, or never resulted in a reward. In such cases,
     this function includes the option to impute the "predictions" for them (which would otherwise
     always be exactly zero or one regardless of the context) by replacing them with random
-    numbers ~Beta(3,1) or ~Beta(1,3) for the cases of always good and always bad.
+    numbers :math:`\sim \text{Beta}(3,1)` or :math:`\sim \text{Beta}(1,3)` for the cases of
+    always good and always bad.
     
     This is just a wild idea though, and doesn't guarantee reasonable results in such siutation.
     
@@ -120,7 +121,7 @@ class DoublyRobustEstimator:
         else:
             assert isinstance(nchoices, int)
             assert nchoices > 2
-            assert ('fit' in dir(base_algorithm)) and ('predict' in dir(base_algorithm))
+            assert hasattr(base_algorithm, "fit") and hasattr(base_algorithm, "predict")
         
         if c is not None:
             assert isinstance(c, float)
@@ -130,9 +131,8 @@ class DoublyRobustEstimator:
         
         if type(reward_estimator) == np.ndarray:
             assert reward_estimator.shape[1] == nchoices
-            assert reward_estimator.shape[0] == X.shape[0]
         else:
-            assert ('predict_proba_separate' in dir(reward_estimator)) or ('predict_proba' in dir(reward_estimator))
+            assert hasattr(reward_estimator, "predict_proba_separate") or hasattr(reward_estimator, "predict_proba")
 
         if beta_prior is not None:
             beta_prior = _check_beta_prior(beta_prior, nchoices, 2)
@@ -167,7 +167,7 @@ class DoublyRobustEstimator:
         """
         try:
             from costsensitive import RegressionOneVsRest, WeightedAllPairs
-        except:
+        except Exception:
             raise ValueError("This functionality requires package 'costsensitive'.\nCan be installed with 'pip install costsensitive'.")
         p = _check_1d_inp(p)
         assert p.shape[0] == X.shape[0]
@@ -175,9 +175,9 @@ class DoublyRobustEstimator:
         
         if type(self.reward_estimator) == np.ndarray:
             C = self.reward_estimator
-        elif 'predict_proba_separate' in dir(self.reward_estimator):
+        elif hasattr(self.reward_estimator, "predict_proba_separate"):
             C = -self.reward_estimator.predict_proba_separate(X)
-        elif 'predict_proba' in dir(self.reward_estimator):
+        elif hasattr(self.reward_estimator, "predict_proba"):
             reward_estimator = \
                 SeparateClassifiers(self.reward_estimator, self.nchoices,
                     beta_prior = self.beta_prior, smoothing = self.smoothing,
@@ -286,7 +286,7 @@ class OffsetTree:
                  random_state = 1, njobs = -1):
         try:
             from costsensitive import _BinTree
-        except:
+        except Exception:
             raise ValueError("This functionality requires package 'costsensitive'.\nCan be installed with 'pip install costsensitive'.")
         _check_constructor_input(base_algorithm, nchoices)
         self.base_algorithm = base_algorithm
@@ -347,7 +347,8 @@ class OffsetTree:
         y_node[r_more_onehalf] = 1. - y[r_more_onehalf]
         w_node = (.5 - r_node) / p_node
         w_node[r_more_onehalf] = (  (r_node - .5) / p_node  )[r_more_onehalf]
-        w_node = w_node * (w_node.shape[0] / np.sum(w_node))
+        if w_node.shape[0] != 0:
+            w_node = w_node * (w_node.shape[0] / np.sum(w_node))
         
         n_pos = y_node.sum()
         if y_node.shape[0] == 0:
